/*
*  File             : build.c 
*  Version          : 1.0
*  Written by       : alinden@netcologne.de
*  Date             : Jul 17 1998
*  Compiler         : gcc 2.7.2 on Linux 2.0.33
*  Portability      : Should be POSIX compliant  
*  Copyright:       : GPL
*
*  DESCRIPTION: Contains the stuff needed to build the crontab database.
*  This is done by calling read_config(). Variable size information from a
*  crontab is stored at the end of each CONFIG struct. The list is concatenated
*  twice for user and entry. To be memory and CPU efficient, the time fields
*  are stored as a bit oriented index within this struct. For example, the minute
*  field is an array of 60 bits in 8 bytes where each bit corresponds to a
*  minute (0-59). A lookup then needs only one access to check if the bit of the
*  current minute is set. 
*/

#include "conf.h"
#include "crond.h"
#include "extern.h"

void msgout(char *); 
void syslogw(char *, int);
char *cpstr(size_t, char *, ...);
void rmenv(NVR *);
time_t runtab(CONFIG *, time_t);
void outfifo(char *);
void flushqueue(void);
void flushqctab(char *);

unsigned long uhours(char **inbuf)   /* returns number of minutes for U option */ 
{                                    /* inbuf: field array */                
	unsigned long tminutes,timeval;

	tminutes = timeval = 0;       /* minute field */
	if ( *inbuf[0] != '*' && sscanf(inbuf[0], "%lu", &timeval) < 1 )
		return(0);
	tminutes = timeval;

	timeval = 0;                 /* hour field */           
	if ( *inbuf[1] != '*' && sscanf(inbuf[1], "%lu" , &timeval ) < 1 )
		return(0);
	tminutes += timeval*60;

	timeval = 0;                 /* day field */
	if ( *inbuf[2] != '*' && sscanf(inbuf[2], "%lu", &timeval) < 1 )
		return(0);
	tminutes += timeval*24*60;

	timeval = 0;                 /* week field */
	if ( *inbuf[3] != '*' && sscanf(inbuf[3], "%lu", &timeval) < 1 )
		return(0);
	tminutes += timeval*24*60*7;
	
	return(tminutes);
}

/* set a bit on a time index field in CONFIG struct. 
   tbuf points to element in CONFIG, bnum = bit number */  

void stfd(unsigned char *tbuf, int bnum)
{
	int cnum;

	if ( bnum < 0 )
		return;

	cnum = bnum / 8;   /* byte number */ 
	tbuf[cnum] |= (0x80 >> (unsigned)(bnum-cnum*8)); 
}	

/* input: string of a time field (*instr) in a crontab entry. Use stfd()
   to convert the string to a bitmap starting at *tbuf.
   max: max. number of bits. Return 0: no bits set, 1: bits are set,
   2: field was unrestricted ('*'), all bits are set  */

int chk_tmstr(char *instr, unsigned char *tbuf, int max)
{
	unsigned rdtm,nrdtm,srdtm;
	int ind;

	for(ind = 0; ind <= max/8; ind++)  /* set bits to zero */
		tbuf[ind] = 0;

	while(*instr)
	{
		while(1)
		{
			if(*instr == '*')
			{
				if(instr[1] == 0 || instr[1] == ',')  /* unrestricted */
				{
					for(ind = 0; ind <= max/8; ind++)  /* set all bits */
						tbuf[ind] = 0xff;

					return(2);	
				}

				if(sscanf(&instr[1], "/%d", &rdtm) < 1 || rdtm == 0)
					break;

				for(ind = 0; ind < max; ind++)	
				{
					if ( ind % rdtm == 0)
						stfd(tbuf, ind);
				}		
				break;
			}

			if(sscanf(instr, "%d-%d", &rdtm, &nrdtm) > 1)
			{
				srdtm = 1;
				ind = strcspn(instr, ",/"); 

				if ( instr[ind] == '/' && (sscanf(&instr[ind+1], "%d", &srdtm) < 1 || srdtm == 0))
						break;

				for(ind = 0; ind < max; ind++)
				{
					if ( ((rdtm <= nrdtm && ind >= rdtm && ind <= nrdtm) || ( rdtm > nrdtm && (ind <= nrdtm || ind >= rdtm))) && (ind % srdtm) == 0)
						stfd(tbuf, ind);
				}		
				break;
			}
	
			if(sscanf(instr, "%d", &rdtm) > 0 && rdtm < max)
				stfd(tbuf, rdtm);
			break;	
		}	
		while( *instr && *(instr++) != ',' );
	}			
	
	for(ind = 0; ind <= max/8; ind++)   /* bits have been set */
	{
		if(tbuf[ind] )
			return(1);
	}		
	return(0);
}  

void lbnm(char *instr, int month)   /* replace labels in string against numbers */
{              	                    /* month=1 : use month table */
	int ind = 0,nind,lst;
	char *outstr;
	const char *ymonth[2][12] = {{ "sun","mon","tue","wed","thu","fri","sat" },{ "jan","feb","mar","apr","may","jun","jul","aug","sep","oct","nov","dec" }};

	( month ) ? ( lst = 12 ) : ( lst = 7 ); 

	outstr = instr;

nloop:	while ( *instr )
	{
		for( nind = 0; nind < lst; nind++)
		{
			ind = 0;
		
			while(1)
			{
				if ( ! instr[ind] || tolower(instr[ind]) != *(ymonth[month][nind]+ind))
					break;
				
				if ( ind++ == 2 )
				{
					sprintf(outstr++, "%d", nind+month);
					if ( nind+month > 9 ) 
						outstr++;
					instr += 3;
					goto nloop;
				}		
			}
		}
		*(outstr++) = *(instr++);
	}
	*outstr = 0;	
} 

int rmuser(char *ctab)  /* remove a user */
{
	CONFIG *tbuf,*pbuf,*puser;
	int nentries=0,suser=0;
	
	if(strcmp(ctab, NCTROOT) == 0)
		ctab = NULL;

	tbuf = pbuf = puser = cblock;

	while(tbuf != NULL)  /* find user */
	{
		if((ctab == NULL && tbuf->cmain) ||
		  (ctab != NULL && ! tbuf->cmain && strcmp(ctab, tbuf->username) == 0))
			break;

		puser = tbuf;
		suser = 1;
		tbuf = tbuf->nuser;
	}
	
	if(tbuf == NULL)
		return(0);
		
	flushqctab(ctab);
	rmenv(tbuf->envr);

	do   /* remove entries for user */
	{
		pbuf = tbuf;
		tbuf = tbuf->nblock;
		free(pbuf);
		nentries++;
		
	} while(tbuf != NULL && tbuf->stblock == 0);

	/* close the chain */

	if(! suser)
		cblock = tbuf;
	else
	{
		puser->nuser = tbuf;

		do
		{
			pbuf = puser;
			puser = puser->nblock;
		
		} while(puser != NULL && puser->stblock == 0);

		pbuf->nblock = tbuf;
	}	
	
	return(nentries);
}

void rmenv(NVR *nvrbuf)  /* remove environment for user */
{
	NVR *pvrbuf;
	
	while(nvrbuf != NULL)
	{
		pvrbuf = nvrbuf;
		nvrbuf = nvrbuf->nblock;
		free(pvrbuf);
	}
}			

int adduser(CONFIG *tbuf)  /* add a user */
{
	CONFIG *pbuf=NULL,*ostart=cblock;
	int nentries=0;
	
	cblock = tbuf;
	tbuf->nuser = ostart;

	while(tbuf != NULL)
	{
		pbuf = tbuf;
		tbuf = tbuf->nblock;
		nentries++;
	}
	
	pbuf->nblock = ostart;

	return(nentries);
}

/* take instr as crontab entry and separate tokens until command is reached
   return array of fields or NULL */

char **sptoken(char *instr) 
{
	static char *ninstr[20];
	int i,i2; 		    			 
			
	if(! *instr)
		return(NULL);
		
	if((ninstr[0] = strtok(instr, "\t ")) == NULL)
		return(NULL);
		
	for(i = 1; i < 6; i++)
	{
		if((ninstr[i] = strtok(NULL, "\t ")) == NULL)
			return(NULL);
	}

	i--;

	while(*ninstr[i] == '-')  /* options field */
	{
		if(i == 19 || (ninstr[++i] = strtok(NULL, "\t ")) == NULL)
			return(NULL);
	}

	ninstr[i+1] = NULL;

	i2 = strcspn(ninstr[i], "\n");
	
	if(*(ninstr[i]+i2) == '\n')
		*(ninstr[i]+i2) = 0;
	else
	{
		*(ninstr[i]+i2) = ' ';
		*(ninstr[i]+strcspn(ninstr[i], "\n")) = 0;   /* chop trailing \n */
	}	

	return(ninstr);
}	

/* check if crontab is owned by user, return -1 if not, 0 if user crontab, 1 if root crontab */

char userperm(char *ctab)
{					       			
	struct passwd *pword;
	char *username,cmain;
	struct stat buf;

	if(strcmp(ctab, NCTROOT) == 0)
	{
		username = "root";
		cmain = 1;
	}
	else
	{
		username = ctab;
		cmain = 0;
	}	
	
	while((pword = getpwnam(username)) == NULL)    /* no such user */
	{
		if(errno != ENOMEM)
			return(-1);
		
		sleep(1);
	}	

	if(stat(ctab, &buf) < 0)
		return(-1);

	if(buf.st_uid && buf.st_uid != pword->pw_uid )  /* not owner of file */
	{	
		syslogw(cpstr(100, "unauthorized crontab: %s", ctab), LOG_INFO);
		return(-1);
	}	
	return(cmain);
}

CONFIG *readconfig(char *ctab, char cmain)  /* read a crontab, return the first block or NULL */
{
	FILE *crontab; 
	char *ninbuf,**cpbuf,dopt,bopt,*outbuf;
	char *inds[3],*nbuf;
	CONFIG *tbuf,*pbuf=NULL;
	static CONFIG *sblock;
	NVR *nvrbuf,*pvrbuf=NULL,*snvr=NULL;
	char inbuf[11*1025];
	size_t slen;
	time_t ltm;
	unsigned long tminutes; 
	int i,dres,wres,nentries=0,ncmds;

	if((crontab = fopen(cpstr(1024, "./%s", ctab), "r")) == NULL)
		return NULL;

	sblock = NULL;

	while(time(&ltm) < 0)
		sleep(1);

	while(fgets(inbuf, 1024, crontab) != NULL)
	{
		do
		{
			for(nbuf = inbuf+strlen(inbuf); nbuf > inbuf && (*nbuf == ' ' || *nbuf == '\t' || *nbuf == '\n' || ! *nbuf); nbuf--);

			if(*nbuf != '\\')  /* no continuation */
				break;

			if(! cmain && nentries++ >= UTABMAX)
				break;

			if(strlen(inbuf) > 10*1024)
				break;

		} while(fgets(nbuf, 1024, crontab) != NULL);
		
		if(strchr(inbuf, '\n') == NULL)
		{
			*(inbuf+(i = strlen(inbuf))) = '\n';
			*(inbuf+i+1) = 0;
		}

		ninbuf = inbuf + strspn(inbuf, "\t ");  /* strip leading space */

		if(*ninbuf == '#' || *ninbuf == '\n')
			continue;

		/* limit user crontabs to 200 entries */
	
		if(! cmain && nentries++ >= UTABMAX)
			break;

		/* test for environment entry */

		outbuf = ninbuf + (i = strcspn(ninbuf, "\t\n= "));

		if (i && *outbuf == '=')
		{
			*(outbuf + strcspn(outbuf, "\n")) = 0;
			while((nvrbuf = malloc(sizeof(NVR)+strlen(ninbuf)+1)) == NULL)
				sleep(1);

			strcpy((char *)nvrbuf+sizeof(NVR), ninbuf);
			nvrbuf->nblock = NULL;
			
			(snvr == NULL) ? (snvr = nvrbuf) : (pvrbuf->nblock = nvrbuf);
			pvrbuf = nvrbuf;
			continue;
		}

		if((cpbuf = sptoken(ninbuf)) == NULL)
			continue;

		inds[0] = (cmain)?("root"):(ctab);
	
		/* make entry of cron field */

		if(*cpbuf[0] == 'U' || *cpbuf[0] == 'u')  /* U option used */ 
		{
			if((tminutes = uhours(&cpbuf[1])) == 0)
				continue;
		}
		else
		{	
			lbnm(cpbuf[3], 1);
			lbnm(cpbuf[4], 0);

			tminutes = 0;
		}	

		/* options */

		ncmds = 5;
		bopt = dopt = 0;
		inds[2] = NULL;

NBLV:		while(*cpbuf[ncmds] == '-')
		{
			for(i = 1; *(cpbuf[ncmds]+i); i++)
			{
				if(tolower(*(cpbuf[ncmds]+i)) == 'u' && *(cpbuf[ncmds]+i+1) == '=')
				{
					if(cmain)
						inds[0] = cpbuf[ncmds]+i+2;

					ncmds++;
					goto NBLV;
				}	
		
				if(! tminutes && tolower(*(cpbuf[ncmds]+i)) == 'd')
					dopt = 1;
				
				if(tolower(*(cpbuf[ncmds]+i)) == 'b')
					bopt = 1;

				if(tolower(*(cpbuf[ncmds]+i)) == 's' && *(cpbuf[ncmds]+i+1) == '=')
				{
					inds[2] = cpbuf[ncmds]+i+2;
				
					ncmds++;
					goto NBLV;
				}
			}	
			ncmds++;
		}

		inds[1] = cpbuf[ncmds];

		/* inds[0] points to username, inds[1] to cmdline , inds[2] to subject now */

		slen = strlen(inds[0])+strlen(inds[1])+2;

		if(inds[2] != NULL)
			slen += strlen(inds[2])+1;

		while((tbuf = malloc(sizeof(CONFIG)+slen)) == NULL)
			sleep(1);

		if(! tminutes)
		{	
			if ( ! chk_tmstr(cpbuf[0], tbuf->minutes, 60)
			  || ! chk_tmstr(cpbuf[1], tbuf->hours, 24)
			  || ! (dres = chk_tmstr(cpbuf[2], tbuf->days, 32))
			  || ! chk_tmstr(cpbuf[3], tbuf->month, 13)
			  || ! (wres = chk_tmstr(cpbuf[4], tbuf->weekday, 7)))
			{
			 	free(tbuf);
			 	continue;
			}

			if( wres == 2 )
				tbuf->wdres = 0x01;
			else
				tbuf->wdres = 0;		

			if( dres == 2 )
				tbuf->wdres |= 0x02;	
		}
			
		/* first pointer into cblock,then expand CONFIG struct */

		(sblock == NULL) ? (sblock = tbuf) : (pbuf->nblock = tbuf); 

		tbuf->username = strcpy((char *)tbuf+sizeof(CONFIG), inds[0]);
		tbuf->cmdstr = strcpy(tbuf->username+strlen(inds[0])+1, inds[1]);

		if(inds[2] != NULL)
			tbuf->jobid = strcpy(tbuf->cmdstr+strlen(inds[1])+1, inds[2]);
		else
			tbuf->jobid = NULL;	
		 
		tbuf->nblock = NULL;
		tbuf->nuser = NULL;
		tbuf->period = tminutes;
		tbuf->dopt = dopt;
		tbuf->bopt = bopt;
		tbuf->stblock = (tbuf == sblock)?(1):(0); 
		tbuf->cmain = cmain;
		tbuf->envr = snvr;
		tbuf->pid = 0;
		tbuf->nextexec = runtab(tbuf, ltm);
		tbuf->lastexec = 0;

		pbuf = tbuf;
	}
	fclose(crontab);

	if(sblock == NULL)
		rmenv(snvr);

	return(sblock);
}

void update_ctab(char *ctab)
{
	char cmain=0;
	CONFIG *tbuf;
	int jobs=0;
	struct stat buf;

	if(strcmp(ctab, NCTROOT) == 0)
		cmain = 1;
	
	if(stat(ctab, &buf))  /* crontab removed  */
	{
		njobs -= rmuser(ctab);
		syslogw(cpstr(100, "removed %s%s", (cmain)?(ctab):("user "), (cmain)?(""):(ctab)), LOG_INFO);

		return;
	}

	if((cmain = userperm(ctab)) == -1)
		return;
	
	njobs -= rmuser(ctab);

	if((tbuf = readconfig(ctab, cmain)) != NULL)
	{
		jobs = adduser(tbuf);
		njobs += jobs;
	}
	syslogw(cpstr(100, "registered %d jobs for %s%s (total %d)", jobs, (cmain)?("root"):("user "),
		(cmain)?(""):(ctab), njobs), LOG_INFO);
}	

void read_config(void)   /* read all crontabs */
{
	DIR *crontabs;
	CONFIG *tbuf,*pbuf;
	int nuser=0,cunr=0;
	char cmain;
	struct dirent *dentry;

	if(chdir(CRONTABS) < 0)
	{
		syslogw("cannot chdir to crontabs: %m", LOG_ERR);
		return;
	}

	if(rwfifo != -1)
		close(rwfifo);

	if(mkfifo("./@control-in", 0600) == -1 && errno != EEXIST)
		syslogw("cannot create fifo: %m", LOG_WARNING);

	if((rwfifo = open("./@control-in", O_RDWR)) == -1)
		syslogw("cannot open fifo: %m", LOG_WARNING);

	if((crontabs = opendir(CRONTABS)) == NULL)
	{
		syslogw("cannot open crontab directory: %m", LOG_ERR);
		return;
	}

	/* remove all crontabs */

	flushqueue();

	tbuf = pbuf = cblock;

	while(tbuf != NULL)
	{
		if(tbuf->stblock)
			rmenv(tbuf->envr);
			
		pbuf = tbuf;
		tbuf = tbuf->nblock;
		free(pbuf);
	}
	cblock = NULL;
	njobs = 0;

	while((dentry = readdir(crontabs)) != NULL)
	{
		if(*dentry->d_name == '.' || strcmp(dentry->d_name, "@control-in") == 0 ||
		strcmp(dentry->d_name, "@cron.restart") == 0)
			continue;
			
		if(strncmp(dentry->d_name, "@output-", 8) == 0)
		{
			unlink(dentry->d_name);
			continue;
		}	

		if((cmain = userperm(dentry->d_name)) == -1)	
		{
			cunr++;
			continue;
		}

		if((tbuf = readconfig(dentry->d_name, cmain)) != NULL)
		{
			njobs += adduser(tbuf);
			nuser++;
		}
	}
	closedir(crontabs);

	syslogw(cpstr(100, "registered %d jobs, %d users", njobs, nuser), LOG_INFO);

	if(cunr)
		syslogw(cpstr(100, "%d files in crontabdir ignored", cunr), LOG_INFO);
}
