/*
*  File                  : crontab.c  
*  Version               : 1.0
*  Last modification by  : alinden@netcologne.de
*  Date                  : Jul 17 1998
*  Compiler              : gcc 2.7.2 on Linux 2.0.33
*  Portability           : Should be POSIX compliant
*  Copyright             : GPL
*  DESCRIPTION:          : Calls set_crontab() to verify permissions, modifies  
*                        : the crontab and sends a command to crond. 
*/

#include "crond.h"
#include "conf.h"

char *msgupd = "Updated";

/* verify permissions and return the crontab name for a user. 
   username = NULL: assume user or NCTROOT if root. */

char *setcrontab(char *username)
{ 
	uid_t userid;
	struct passwd *pword,*upword;

	if((pword = getpwuid(userid = getuid())) == NULL)
		exit(1);

	if(userid == 0)  /* user is root */
	{
		if(username == NULL || strcmp(username, NCTROOT) == 0)
			return(NCTROOT);

		/* Verify if user does exist */	

		if((upword = getpwnam(username)) != NULL)
			return(upword->pw_name);
			
		fputs("Unknown user\n", stderr);
		exit(1);	
	}	

	/* not root */

	if(username == NULL )  /* no name specified */
		return(pword->pw_name);  /* own name */
		
	if((upword = getpwnam(username)) == NULL || pword->pw_uid != upword->pw_uid )
	{
		fputs("Only root will specify crontabs of other users\n", stderr);
		exit(1);
	}	
	
	return(pword->pw_name);
}

char *set_crontab(char *username)   /* call setcrontab and add preceding ./ to crontab */  
{
	char *crontab;
	static char buf[203] = { '.','/' };

	buf[202] = 0;
	crontab = setcrontab(username);
	strncpy(&buf[2], crontab, 200);

	return(buf);
}

void catf(int infile, int outfile)    /* cat file from infile to outfile */
{
	char buf[1024];
	ssize_t inbytes;

	while((inbytes = read(infile, buf, 1024)) > 0)
		write(outfile, buf, inbytes);  	 	 
}   

void cdir_ctab(void)   /* change to the crontab dir */
{
	if(chdir(CRONTABS) < 0)
	{
		perror("Cannot access crontab directory");
		exit(1);
	}	
}

void sig_alrm(int signo) { return; }

int send_signal(char *ctab)   /* send command to pipe  */
{
	FILE *rwfifo;
	struct stat buf;
	int fd,i;

	signal(SIGALRM, (void *)sig_alrm);
	alarm(6);
	fd = open("./@control-in", O_WRONLY);
	alarm(0);
	
	if(fd  == -1 || (rwfifo = fdopen(fd, "w")) == NULL)
	{
		fprintf(stderr, "Cannot write to the pipe. No cron running???\n");
		return(0);
	}

	flock(fd, LOCK_EX);

	for(i = 0; fstat(fd, &buf) == 0 && buf.st_size; i++)
	{
		if(i > 12)
		{
			fputs("Pipe busy\n", stderr);
			return(0);
		}
		sleep(1);
	}	

	fprintf(rwfifo, "%s\n", ctab+2);
	
	fflush(rwfifo);
	flock(fd, LOCK_UN);
	fclose(rwfifo);

	return(1);
}		

int lcrontab(char *crontab)     /* list a crontab */
{
	int ctab;

	cdir_ctab();

	if ( (ctab = open(crontab, O_RDONLY)) < 0)
	{
		perror("Cannot open this crontab");
		exit(1);
	}

	catf(ctab, 1);
	
	close(ctab);

	exit(0);
}

int dcrontab(char *crontab)  /* delete a crontab */
{
	struct stat buf;

	cdir_ctab();

	if(lstat(crontab, &buf) == 0 && S_ISLNK(buf.st_mode))
	{
		fprintf(stderr, "Crontab is a symbolic link\n");
		exit(1);
	}	

	if(unlink(crontab))
	{
		perror("No crontab deleted");
		exit(1);
	}	

	if(send_signal(crontab))
		puts(msgupd);
	
	exit(0);	
}

int ecrontab(char *crontab)  /* edit a crontab */
{
	pid_t pid;
	uid_t userid;
	gid_t groupid;
	char *editor,*neditor;
	struct stat buf;
	time_t lmtime;
	int ctab;

	cdir_ctab();

	if((ctab = open(crontab, O_RDONLY | O_CREAT, 0644)) < 0)
	{
		perror("Cannot open crontab");
		exit(1);
	}	

	groupid = getgid();

	if((userid = getuid()))    /* set crontab to current user */
		fchown(ctab, userid, groupid);

	/* get the last modification time to verify that it was changed */

	fstat(ctab, &buf); 
	lmtime = buf.st_mtime;
	
	close(ctab);

	if((pid = fork()) < 0 )
	{
		perror("Fork");
		exit(1);
	}
	 
	if(pid == 0)   /* child */
	{
		if(userid && (setgid(groupid) || setuid(userid)))
		{
			perror("Cannot set permissions");
			exit(1);
		}

		/* verify environment to get editor name */

		if(((editor = getenv("VISUAL")) == NULL && (editor = getenv("EDITOR")) == NULL) || *editor == 0)
			editor = EDITOR;

		if((neditor = strrchr(editor, '/')) == NULL)
			neditor = editor;
		else
			neditor++;

		if(execlp(editor, neditor, crontab, NULL) < 0)
		{
			perror("Cannot run the editor"); 
			fprintf(stderr, "Editor name: %s\n", editor);
		}
		exit(0);
	}

	wait(NULL);

	stat(crontab, &buf); 

	if ( buf.st_mtime == lmtime )  /* File left unchanged */
		exit(0);

	if(send_signal(crontab))
		puts(msgupd);

	exit(0);
}	

int reqinfo(char *crontab, char *rexpr)
{
	char cbuf[190],cmdbuf[198];
	int infifo;
	
	cdir_ctab();

	if(snprintf(cbuf, 190, "./@output-%s", crontab+2) == -1)
		exit(1);

	if(mkfifo(cbuf, 0600) == -1 && errno != EEXIST)
	{
		perror("Cannot create pipe for reading");
		exit(1);
	}

	if(*rexpr == '*')
	{
		snprintf(cmdbuf, 198, "%s @*", cbuf);
		
	} else if(snprintf(cmdbuf, 198, "%s @*%s*", cbuf ,rexpr) == -1)
	{
		fputs("Command too long\n", stderr);
		exit(1);
	}

	if(send_signal(cmdbuf) == 0)
		exit(1);

	signal(SIGALRM, (void *)sig_alrm);
	alarm(6);

	infifo = open(cbuf, O_RDONLY);

	alarm(0);

	if(infifo == -1)
	{
		if(errno == EINTR)
			fprintf(stderr, "Pipe timed out (no reply received)\n");
		else
			perror("Cannot open input pipe");
			
		unlink(cbuf);
		exit(1);
	}

	catf(infifo, 1);
	unlink(cbuf);

	exit(0);
}

int runproc(char *crontab, char *rcmd)
{
	char cbuf[190],cmdbuf[198];
	int infifo;
	
	cdir_ctab();

	if(snprintf(cbuf, 190, "./@output-%s", crontab+2) == -1)
		exit(1);

	if(mkfifo(cbuf, 0600) == -1 && errno != EEXIST)
	{
		perror("Cannot create pipe for reading");
		exit(1);
	}

		
	if(snprintf(cmdbuf, 198, "%s @R%s", cbuf ,rcmd) == -1)
	{
		fputs("Command too long\n", stderr);
		exit(1);
	}

	if(send_signal(cmdbuf) == 0)
		exit(1);

	signal(SIGALRM, (void *)sig_alrm);
	alarm(6);

	infifo = open(cbuf, O_RDONLY);

	alarm(0);

	if(infifo == -1)
	{
		if(errno == EINTR)
			fprintf(stderr, "Pipe timed out (no reply received)\n");
		else
			perror("Cannot open input pipe");

		unlink(cbuf);
		exit(1);
	}
	
	catf(infifo, 1);
	unlink(cbuf);

	exit(0);
}

int fcrontab(char *crontab, char *infile)  /* replace crontab from file */
{
	uid_t userid;
	gid_t groupid;
	int intab, outtab;

	if(*infile == '-' && ! *(infile+1))
		intab = 0;
	else
	{
		if(access(infile, R_OK) || (intab = open(infile, O_RDONLY)) == -1)
		{
			perror("Cannot open input file");
			exit(1);
		}
	}	

	cdir_ctab();

	if((outtab = open(crontab, O_WRONLY | O_CREAT | O_TRUNC, 0644)) < 0)
	{
		perror("Cannot open crontab");
		exit(1);
	}
	
	groupid = getgid();

	if((userid = getuid()))   /* set owner to current user */
		chown(crontab, userid, groupid);

	catf(intab, outtab);

	close(intab);
	close(outtab);
	
	if(send_signal(crontab))
		puts(msgupd);
	
	exit(0);
}			

int main(int argc, char *argv[])
{
	int i;
	char *username=NULL,*filename=NULL,*rexpr=NULL,*rcmd=NULL;
	void (*func)(char *);

	func = NULL;
	
	umask(066);

	if(argc > 1)
	{	
		if(strcmp(argv[1], "--version") == 0)
		{
			puts(VERSION);
			exit(0);
		}

		for(i = 1; i < argc; i++)
		{
			if(*argv[i] != '-')
			{
				filename=argv[i]; 
				continue;
			}	

			switch(*(argv[i]+1))
			{
				case 'u':
						if(++i < argc)
							username = argv[i];
						break;	
				case 'l':
						func = (void *)lcrontab;
						break;
				case 'd':
						func = (void *)dcrontab;
						break;
				case 'e':
						func = (void *)ecrontab;
						break;
				case 'i':
						if(i+1 < argc)
						{
							if(*argv[i+1] == '-')
							{
								rexpr = "*";
								break;
							}	
							
							rexpr = argv[++i];
						}
						else
							rexpr = "*";
						break;
				case 'r':
						if(++i < argc)
							rcmd = argv[i];
						break;
				case 0:
						filename = "-";
			}
		}
		if(rexpr != NULL)
			reqinfo(set_crontab(username), rexpr);

		if(rcmd != NULL)
			runproc(set_crontab(username), rcmd);
		
		if(func == NULL && filename != NULL)
			fcrontab(set_crontab(username), filename);

		if(func != NULL && filename == NULL)	
			(*func)(set_crontab(username));
	}
	fputs("Usage: crontab [-u <user>] { <file>|-e|-d|-l|-i [<pattern>]|-r <cmd> }\n", stderr);
	exit(1);
}	

